import math

from pdlpy.distribution import Distribution


class Normal(Distribution):
    """
    Continuous probability distribution of the random variable X that is assumed to be additively produced by many small effects
    """

    def __init__(self, mean, var):
        """
        Paramters
        mean: the expectation of the distribution
        var: the variance of the distribution
        """
        self._mean = mean
        self._var = var

    def __str__(self):
        mean = round(self._mean, 2)
        var = round(self._var, 2)
        return f"Normal(mean={mean}, var={var})"

    def pdf(self, x):
        """
        Probability Density Function

        Paramters
        x: a value of random variable X

        Returns
        the relative likelihood that a value of X would lie in sample space
        """
        return (1 / math.sqrt(2 * math.pi * self._var)) * math.e ** (
            -((x - self._mean) ** 2 / 2 * self._var)
        )

    def cdf(self, x):
        """
        Cumulative Distribution Function

        Parameters
        x: a value of the random variable X

        Returns
        the probability that X will take a value less than or equal to x
        """
        return (1 + math.erf((x - self._mean) / (math.sqrt(self._var * 2)))) / 2



from pdlpy.distribution import Distribution


class Uniform(Distribution):
    """
    Continuous distribution of a random variable X in interval [a; b] where any value of X has an equal probability
    """

    def __init__(self, a, b):
        """
        Paramters
        a: the minimum value of X
        b: the maximum value of X
        """
        self._a = a
        self._b = b
        self._mean = (a + b) / 2
        self._var = (b - a) ** 2 / 12

    def __str__(self):
        a = round(self._a, 2)
        b = round(self._b, 2)
        mean = round(self._mean, 2)
        var = round(self._var, 2)
        return f"Uniform(a={a}, b={b}, mean={mean}, var={var})"

    def pdf(self, x=None):
        """
        Probability Density Function

        Paramters
        x: a value of random variable X

        Returns
        the relative likelihood that a value of X would lie in sample space
        """
        return 1 / (self._b - self._a)

    def cdf(self, x):
        """
        Cumulative Distribution Function

        Parameters
        x: a value of the random variable X

        Returns
        the probability that X will take a value less than or equal to x
        """
        if x <= self._a:
            return 0.0
        elif x >= self._b:
            return 1.0
        else:
            return (x - self._a) / (self._b - self._a)



from pdlpy.combinatorics import ncr
from pdlpy.distribution import Distribution


class Hypergeometric(Distribution):
    """
    Discrete probability distribution that describes the probability of successes in draws from a finite test set
    """

    def __init__(self, n, N, M):
        """
        Parameters
        n: the number of draws
        N: the size of the test set
        M: the number of successes
        """
        self._n = n
        self._N = N
        self._M = M
        self._mean = n * M / N
        self._var = n * M / N * (1 - M / N) * (N - n) / (N - 1)

    def __str__(self):
        n = round(self._n, 2)
        N = round(self._N, 2)
        M = round(self._M, 2)
        mean = round(self._mean, 2)
        var = round(self._var, 2)
        return f"Hypergeometric(n={n}, N={N}, M={M}, mean={mean}, var={var})"

    def pmf(self, x):
        """
        Probability Mass Function

        Parameters
        x: a value of the random variable X

        Returns
        the probability that X will take a value exactly equal to x
        """
        return (
            ncr(self._M, x)
            * ncr(self._N - self._M, self._n - x)
            / ncr(self._N, self._n)
        )

    def cdf(self, x):
        """
        Cumulative Distribution Function

        Parameters
        x: a value of the random variable X

        Returns
        the probability that X will take a value less than or equal to x
        """
        if x == 0:
            return self.pmf(0)
        else:
            return self.pmf(x) + self.cdf(x - 1)



import math

from pdlpy.distribution import Distribution


class Exponential(Distribution):
    """
    Continuous probability distribution of time between events in a Poisson process
    """

    def __init__(self, rate):
        """
        Paramters
        rate: the average number of events
        """
        self._rate = rate
        self._mean = rate ** -1
        self._var = rate ** -2

    def __str__(self):
        rate = round(self._rate, 2)
        mean = round(self._mean, 2)
        var = round(self._var, 2)
        return f"Exponential(rate={rate}, mean={mean}, var={var})"

    def pdf(self, x):
        """
        Probability Density Function

        Paramters
        x: a value of random variable X

        Returns
        the relative likelihood that a value of X would lie in sample space
        """
        return self._rate * math.e ** (-self._rate * x)

    def cdf(self, x):
        """
        Cumulative Distribution Function

        Parameters
        x: a value of the random variable X

        Returns
        the probability that X will take a value less than or equal to x
        """
        return 1 - math.e ** (-self._rate * x)



import math

from pdlpy.distribution import Distribution


class Poisson(Distribution):
    """
    Discrete probability distribution that expresses the probability of a given number of events occurring in a fixed interval of time or space
    """

    def __init__(self, rate):
        """
        Parameters
        rate: the average number of events
        """
        self._rate = rate
        self._mean = rate
        self._var = rate

    def __str__(self):
        rate = round(self._rate, 2)
        mean = round(self._mean, 2)
        var = round(self._var, 2)
        return f"Poisson(rate={rate}, mean={mean}, var={var})"

    def pmf(self, x):
        """
        Probability Mass Function

        Parameters
        x: a value of the random variable X

        Returns
        the probability that X will take a value exactly equal to x
        """
        return (self._rate ** x) * (math.e ** (-self._rate)) / math.factorial(x)

    def cdf(self, x):
        """
        Cumulative Distribution Function

        Parameters
        x: a value of the random variable X

        Returns
        the probability that X will take a value less than or equal to x
        """
        if x == 0:
            return self.pmf(0)
        else:
            return self.pmf(x) + self.cdf(x - 1)
